/***************************************************************************
                          interpreter.cc  -  description
                             -------------------
    begin                : Tue Mar 21 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdlib.h>
#include "interpreter.h"
#include "output.h"
#include "type.h"

bool parsingbias = true;

bool interpret_type ( int size, AtomString **params ) {
  char *type_name = params[0]->name;
  if ( size < 3 )
    return false;
  typetable.addType ( type_name, size - 1 );
  for ( int i = 1; i < size; i++ ) {
    if ( params[i]->nrofparams > 0 )
      return false;
    typetable.addConstant ( params[i]->name, type_name );
  }
  return true;
}

bool interpret_inttype ( int size, AtomString **params ) {
  if ( size != 2 )
    return false;
  char *type_name = params[0]->name;
  int n = atoi ( params[1]->name );
  typetable.addType ( type_name, n );
  return true;
}

bool interpret_predicate ( int size, AtomString **params ) {
  if ( size != 1 )
    return false;
  char *name = params[0]->name;
  size = params[0]->nrofparams;
  int *paramtypes = new int[size], typenr, nrofconst;
  params = params[0]->params;
  for ( int i = 0; i < size; i++ ) {
    if ( params[i]->nrofparams != 0 )
      return false;
    typetable.getType ( params[i]->name, typenr, nrofconst );
    paramtypes[i] = typenr;
  }
  predicatetable.addPredicate ( name, size, paramtypes );
  
  return true;
}

bool interpret_mode ( int size, AtomString **params, char *functionname = NULL ) {
  if ( size != 2 && size != 3 )
    return false;
  int number = atoi ( params[0]->name ), typenr, constantnr, idnr;
  char *name = params[1]->name;
  if ( size == 3 )
    idnr = atoi ( params[2]->name );
  else
    idnr = -1;
  PredicateData *predicatedata = predicatetable.getPredicate ( name );
  if ( !predicatedata ) {
    ERR << "Unknown predicate " << name << ": ";
    return false;
  }
  
  int predicate = predicatedata->nr,
      predsize = predicatedata->size;
  if ( params[1]->nrofparams != predsize ) {
    ERR << "Wrong number of parameters: ";
    return false;
  }
  ModeParamData *modeparams = new ModeParamData[predsize];
  AtomString **walk = params[1]->params;

  for ( int i = 0; i < predsize; i++ )
    switch ( walk[i]->name[0] ) {
      case '+': {
        modeparams[i].type = Plus;
        modeparams[i].number = predicatedata->params[i];
        break;
      }
      case '-': {
        modeparams[i].type = Min;
	modeparams[i].number = predicatedata->params[i];
	break;
      }
      case '\\': {
        modeparams[i].type = MinEx;
        modeparams[i].number = predicatedata->params[i];
        break;
      }
      default: {
        modeparams[i].type = Constant;
        typenr = predicatedata->params[i];
          // define the type expected, could be used in the
          // search for the number constantnr
	typetable.getConstant ( walk[i]->name, typenr, constantnr );
	if ( constantnr == -1 ) {
	  ERR << "Unknown constant " << walk[i]->name << ": ";
	  return false;
	}
	if ( typenr != -1 && typenr != predicatedata->params[i] ) {
	  ERR << "Constant " << walk[i]->name << " is of wrong type: ";
	  return false;
	}
	modeparams[i].number = constantnr;
      }
    }

  modetable.addMode ( number, predicate, predsize, modeparams, idnr, functionname );
  
  return true;
}

bool interpret_fact ( char *name, int size, AtomString **params ) {
  PredicateData *predicatedata;
  if ( name[0] <= '9' )
    predicatedata = predicatetable.getPredicate ( atoi ( name ) );
  else
    predicatedata = predicatetable.getPredicate ( name );
  int typenr, constantnr;

  if ( !predicatedata ) {
    ERR << "Unknown predicate used in fact: ";
    return false;
  }
  if ( size != predicatedata->size ) {
    ERR << "Incorrect number of parameters: ";
    return false;
  }
  if ( predicatedata->key ) {
    ERR << "Can not use facts for the key: ";
    return false;
  }
  INTTYPE coor[size];
  for ( int i = 0; i < size; i++ ) {
    typenr = predicatedata->params[i];
    typetable.getConstant ( params[i]->name, typenr, constantnr );
    if ( typenr != -1 && ( typenr != predicatedata->params[i] || constantnr == -1 ) ) {
      ERR << "Constant " << params[i]->name << " is of incorrect type: ";
      return false;
    }
    coor[i] = constantnr;
  }
  setFact ( predicatedata, coor );
  return true;
}

bool interpret_key ( int size, AtomString **params ) {
  if ( size != 1 ) {
    ERR << "Too much parameters for key: ";
    return false;
  }
  PredicateData *predicatedata = predicatetable.getPredicate ( params[0]->name );
  if ( !predicatedata ) {
    ERR << "Unknown predicate " << params[0]->name << ": ";
    return false;
  }
  predicatedata->key = true; // not null, not to be created

  // add as mode declaration:
  int predsize = predicatedata->size;
  ModeParamData *modeparams = new ModeParamData[predsize];

  for ( int i = 0; i < predsize; i++ ) {
    modeparams[i].type = Min;
    modeparams[i].number = predicatedata->params[i];
  }

  key = modetable.addMode ( 1, predicatedata->nr, predsize, modeparams, -1, NULL );
  
  return true;
}

bool interpret_minsup ( int size, AtomString **params ) {
  if ( size != 1 ) {
    ERR << "Minimum support needs one integer parameter: ";
    return false;
  }
  minsup = atoi ( params[0]->name );
  return true;
}

bool interpret_minint ( int size, AtomString **params ) {
  if ( size != 1 ) {
    ERR << "Minimum relative interest needs one integer parameter: ";
    return false;
  }
  minint = atof ( params[0]->name );
  return true;
}

bool interpret_maxloop ( int size, AtomString **params ) {
  if ( size != 1 ) {
    ERR << "Maximum loop number needs one integer parameter: ";
    return false;
  }
  maxloop = atoi ( params[0]->name );
  return true;
}

bool interpret_selfduplication ( int size, AtomString **params ) {
  if ( size != 1 ) {
    ERR << "Self duplication needs one parameter: ";
    return false;
  }
  if ( strcmp ( params[0]->name, "true" ) == 0 )
    selfduplication = true;
  else
    if ( strcmp ( params[0]->name, "false" ) == 0 )
      selfduplication = false;
    else {
      ERR << "Self duplication parameter should be true or false ";
      return false;
    }
  return true;
}

bool interpret_showconstantnames ( int size, AtomString **params ) {
  if ( size != 1 ) {
    ERR << "Show constant names needs one parameter: ";
    return false;
  }
  if ( strcmp ( params[0]->name, "true" ) == 0 )
    showconstantnames = true;
  else
    if ( strcmp ( params[0]->name, "false" ) == 0 )
      showconstantnames = false;
    else {
      ERR << "Show constant names parameter should be true or false ";
      return false;
    }
  return true;
}

bool interpret_numberimm ( int size, AtomString **params ) {
  if ( size != 1 ) {
    ERR << "Numberimm needs one parameter: ";
    return false;
  }
  if ( strcmp ( params[0]->name, "true" ) == 0 )
    typetable.set_number_imm ( true );
  else
    if ( strcmp ( params[0]->name, "false" ) == 0 )
      typetable.set_number_imm ( false );
    else {
      ERR << "Numberimm parameter should be true or false ";
      return false;
    }
  return true;
}

void fill () {
  ERR << "[Processing bias]" << endl;
  ERR_F
  typetable.fillArray ();
  predicatetable.fillArray ();
  modetable.fillArray ( predicatetable );
  modetable.fillDependencies ( predicatetable, typetable );
}

bool interpret ( AtomString *atom ) {
  if ( parsingbias ) {
    if ( strcmp ( atom->name, "type" ) == 0 )
      return interpret_type ( atom->nrofparams, atom->params );
     // an inttype does not introduce typenames, just tells
     // the maximum number. Displayed are only numbers,
     // facts should use numbers, etc.
    if ( strcmp ( atom->name, "inttype" ) == 0 )
      return interpret_inttype ( atom->nrofparams, atom->params );
    if ( strcmp ( atom->name, "predicate" ) == 0 )
      return interpret_predicate ( atom->nrofparams, atom->params );
    if ( strcmp ( atom->name, "mode" ) == 0 )
      return interpret_mode ( atom->nrofparams, atom->params );
    if ( strcmp ( atom->name, "externmode" ) == 0 )
      return interpret_mode ( atom->nrofparams, atom->params, atom->nrofparams >= 2 ? atom->params[1]->name : (char*) NULL );
    if ( strcmp ( atom->name, "externnamemode" ) == 0 )
      return interpret_mode ( atom->nrofparams - 1, atom->params + 1, atom->params[0]->name );
    if ( strcmp ( atom->name, "key" ) == 0 )
      return interpret_key ( atom->nrofparams, atom->params );
    if ( strcmp ( atom->name, "minsup" ) == 0 )
      return interpret_minsup ( atom->nrofparams, atom->params );
    if ( strcmp ( atom->name, "minint" ) == 0 )
      return interpret_minint ( atom->nrofparams, atom->params );
    if ( strcmp ( atom->name, "maxloop" ) == 0 )
      return interpret_maxloop ( atom->nrofparams, atom->params );
    if ( strcmp ( atom->name, "numberimm" ) == 0 )
      return interpret_numberimm ( atom->nrofparams, atom->params );
    if ( atom->name[0] == 's' ) { // speedup?
      if ( strcmp ( atom->name, "selfduplication" ) == 0 )
        return interpret_selfduplication ( atom->nrofparams, atom->params );
      if ( strcmp ( atom->name, "showconstantnames" ) == 0 )
        return interpret_showconstantnames ( atom->nrofparams, atom->params );
    }
    parsingbias = false;
      // no bias definition, going to background knowledge.
      // first put bias in faster access datastructures
    fill ();
    ERR << "[Initializing fact tables]" << endl;
    ERR_F
  }

  return interpret_fact ( atom->name, atom->nrofparams, atom->params );
}

void assertfilledbias () {
  if ( parsingbias )
    fill ();
}



